"""Common utils for testing"""

import os
import re
from pathlib import Path

import pytest


def get_files(dir, sub_str=None):
    pattern = "*" if sub_str is None else f"*{sub_str}*"
    return [path for path in dir.rglob(pattern) if path.is_file()]


def update_mtime(path_file):
    with path_file.open("a") as filetoupdate:
        filetoupdate.write("I have been updated.")


def escape_ansi(string):
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", string)


def escape_rich_style(string):
    style_escape = re.compile(r"(?<!\\)\[[^\]]*[^\\]\]")
    return style_escape.sub("", string)


def check_mtime_consistency(file1, file2):
    assert (
        abs(os.path.getmtime(file1) - os.path.getmtime(file2)) <= 1
    ), f"File modification time is not consistent for file {file1}"


def assert_files_present(dir_source, dir_ref, check_mtime=True):
    print("---------------------")
    print(
        f"checking consistencies of folders:"
        f"\n{str(dir_source)}\nand\n{str(dir_ref)}."
    )
    list_files_source = [x for x in dir_source.rglob("*") if not x.is_dir()]
    for cur_file_source in list_files_source:
        cur_file_dest = Path(
            str(cur_file_source).replace(str(dir_source), str(dir_ref))
        )
        print(f"checking file: {cur_file_source}")
        # check for existence
        assert (
            cur_file_dest.exists()
        ), f"file {cur_file_source.name} from {dir_source} not existent in {dir_ref}."
        if check_mtime:
            # check for proper updating
            check_mtime_consistency(cur_file_source, cur_file_dest)
    print("---------------------")


def substr_in_dir_names(directory, sub_str="DELME", files_only=False, dirs_only=False):
    if files_only:
        return any([sub_str in x.name for x in directory.rglob("*") if x.is_file()])
    elif dirs_only:
        return any([sub_str in x.name for x in directory.rglob("*") if x.is_dir()])
    else:
        return any([sub_str in x.name for x in directory.rglob("*")])


def parametrize_default_ids(argname, argvalues, indirect=False, ids=None, scope=None):
    if not ids:
        argnames = argname.split(",") if isinstance(argname, str) else argname
        if len(argnames) > 1:
            ids = [
                "-".join(f"{k}={v}" for k, v in zip(argnames, p_argvalues))
                for p_argvalues in argvalues
            ]
        else:
            ids = [f"{argnames[0]}={v}" for v in argvalues]
    return pytest.mark.parametrize(
        argname, argvalues, indirect=indirect, ids=ids, scope=scope
    )



"""Clifs plugin to create data backups"""

import csv
import shutil
import sys
import time
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Tuple

from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TaskID
from rich.table import Table

from clifs import ClifsPlugin
from clifs.utils_cli import (
    CONSOLE,
    get_count_progress,
    get_last_action_progress,
    print_line,
    set_style,
)


class DirPair(NamedTuple):
    """Source/Destination directory pair"""

    source: Path
    dest: Path


def conditional_copy(
    path_source: Path, path_dest: Path, dry_run: bool = False
) -> Optional[str]:
    """
    Copy only if dest file does not exist or is older than the source file.
    """
    process = None
    if not path_dest.exists():
        process = "adding"
    elif (path_source.stat().st_mtime - path_dest.stat().st_mtime) > 1:
        process = "updating"

    if process is not None and not dry_run:
        path_dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(path_source, path_dest)
    return process


def conditional_delete(
    path_source: Path, path_dest: Path, list_source: List[Path], dry_run: bool = False
) -> int:
    """
    Delete only if `path_source` is not in `list_source`.
    """
    if path_source not in list_source and path_dest.exists():
        if path_dest.is_dir():
            if not dry_run:
                shutil.rmtree(str(path_dest))
        else:
            if not dry_run:
                path_dest.unlink()
        return 1
    return 0


def list_filedirs(dir_source: Path) -> Tuple[List[Path], List[Path]]:
    """
    List files and directories in a source dir.
    """

    list_files = []
    list_dirs = []

    for cur_file in dir_source.rglob("*"):
        if cur_file.is_dir():
            list_dirs.append(cur_file)
        else:
            list_files.append(cur_file)

    return list_files, list_dirs


class FileSaver(ClifsPlugin):
    """
    Create backups from folders.
    """

    dir_source: Optional[Path]
    dir_dest: Optional[Path]
    cfg_file: Optional[Path]
    delete: bool
    verbose: bool
    dry_run: bool

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs plugins.
        """

        parser.add_argument(
            "-s", "--dir_source", type=Path, default=None, help="Source directory"
        )
        parser.add_argument(
            "-d", "--dir_dest", type=Path, default=None, help="Destination directory"
        )
        parser.add_argument(
            "-cfg",
            "--cfg_file",
            type=Path,
            default=None,
            help="Path to the config file",
        )
        parser.add_argument(
            "-del",
            "--delete",
            action="store_true",
            default=False,
            help=(
                "Delete files which exist in destination directory but not in "
                "the source directory"
            ),
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help="Report on every action to stdout.",
        )
        parser.add_argument(
            "-dr",
            "--dry_run",
            action="store_true",
            default=False,
            help="Do not touch anything.",
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)
        self.console = CONSOLE

        if self.cfg_file and self.dir_source or self.cfg_file and self.dir_dest:
            self.console.print(
                "Paths provided in config table and as parameters. "
                "You'll have to decide for one option I am afraid."
            )
            sys.exit(0)

        if self.cfg_file:
            # TODO: check_cfg_format(cfg_file)  # pylint: disable=fixme
            self.dir_pairs = []
            with self.cfg_file.open(newline="\n", encoding="utf-8") as cfg_file:
                reader = csv.DictReader(cfg_file, fieldnames=["source_dir", "dest_dir"])
                # skip header row
                next(reader)
                for row in reader:
                    self.dir_pairs.append(
                        DirPair(Path(row["source_dir"]), Path(row["dest_dir"]))
                    )

        elif self.dir_source and self.dir_dest:
            self.dir_pairs = [DirPair(self.dir_source, self.dir_dest)]

    def run(self) -> None:
        """
        Running the plugin. Needed for all clifs plugins.
        """
        time_start = time.time()

        for dir_pair in self.dir_pairs:
            self.backup_dir(dir_pair.source, dir_pair.dest)
        time_end = time.time()
        time_run = (time_end - time_start) / 60
        self.console.print(
            f"Hurray! All files backed up in only {time_run:5.2f} minutes"
        )

    def get_backup_tasks(
        self, progress: Dict[str, Progress], files_total: int
    ) -> Dict[str, TaskID]:
        return {
            "progress_backup": progress["overall"].add_task(
                "Storing data:  ", total=files_total, last_action="-"
            ),
            "count_files_found": progress["counts"].add_task(
                "Files processed:", total=None
            ),
            "count_files_added": progress["counts"].add_task(
                "Files added:", total=None
            ),
            "count_files_updated": progress["counts"].add_task(
                "Files updated:", total=None
            ),
            "count_files_untouched": progress["counts"].add_task(
                "Files untouched:", total=None
            ),
        }

    def get_delete_tasks(
        self, progress: Dict[str, Progress], files_total: int, dirs_total: int
    ) -> Dict[str, TaskID]:
        return {
            "progress_delete_files": progress["overall"].add_task(
                "Deleting files:",
                total=files_total,
                last_action="-",
                last_action_desc="Last file deleted",
            ),
            "progress_delete_folders": progress["overall"].add_task(
                "Deleting dirs:",
                total=dirs_total,
                last_action="-",
                last_action_desc="Last dir deleted",
            ),
            "count_files_found": progress["counts"].add_task(
                "Files processed:", total=None
            ),
            "count_files_deleted": progress["counts"].add_task(
                "Files deleted:", total=None
            ),
            "count_folders_found": progress["counts"].add_task(
                "Dirs processed:", total=None
            ),
            "count_folders_deleted": progress["counts"].add_task(
                "Dirs deleted:", total=None
            ),
        }

    def backup_dir(
        self,
        dir_source: Path,
        dir_dest: Path,
    ) -> None:
        print_line(console=self.console)
        self.console.print(f"Backing up files \nfrom: {dir_source}\nto:   {dir_dest}")

        if not dir_source.is_dir():
            self.console.print(
                set_style(
                    f"Warning: the source directory does not exist. "
                    f"Nothing to back up from:\n{dir_source}",
                    "warning",
                )
            )
            return

        files_source, dirs_source = list_filedirs(dir_source)
        self.copy_data(
            dir_source=dir_source,
            dir_dest=dir_dest,
            files_source=files_source,
        )

        if self.delete:
            self.console.print("All files stored, checking for files to delete now.")

            self.delete_obsolete_data(
                dir_source=dir_source,
                dir_dest=dir_dest,
                files_source=files_source,
                dirs_source=dirs_source,
            )
        print_line(console=self.console)

    def copy_data(
        self, dir_source: Path, dir_dest: Path, files_source: List[Path]
    ) -> None:
        progress: Dict[str, Progress] = {
            "counts": get_count_progress(),
            "overall": get_last_action_progress(),
        }

        tasks = self.get_backup_tasks(progress, len(files_source))

        progress_table = Table.grid()
        progress_table.add_row(
            Panel.fit(
                progress["overall"],
                title="Progress Backup",
                border_style="cyan",
                padding=(2, 2),
            ),
            Panel.fit(
                progress["counts"],
                title="Counts",
                border_style="bright_black",
                padding=(1, 2),
            ),
        )

        with Live(
            progress_table,
            console=self.console,
            auto_refresh=False,
        ) as live:
            for cur_file in files_source:
                progress["counts"].advance(tasks["count_files_found"])
                action = conditional_copy(
                    cur_file,
                    Path(str(cur_file).replace(str(dir_source), str(dir_dest))),
                    dry_run=self.dry_run,
                )
                if action:
                    progress["overall"].update(
                        tasks["progress_backup"],
                        last_action=f"{action} {cur_file.name}",
                    )
                    if action == "adding":
                        progress["counts"].advance(tasks["count_files_added"])
                    elif action == "updating":
                        progress["counts"].advance(tasks["count_files_updated"])
                    if self.verbose:
                        live.console.print(f"  - {action} [cyan]'{cur_file.name}'[/]")
                else:
                    progress["counts"].advance(tasks["count_files_untouched"])
                progress["overall"].advance(tasks["progress_backup"])
                live.refresh()

    def delete_obsolete_data(
        self,
        dir_source: Path,
        dir_dest: Path,
        files_source: List[Path],
        dirs_source: List[Path],
    ) -> None:
        files_dest, dirs_dest = list_filedirs(dir_dest)

        progress: Dict[str, Progress] = {
            "counts": get_count_progress(),
            "overall": get_last_action_progress(),
        }
        tasks_delete = self.get_delete_tasks(progress, len(files_dest), len(dirs_dest))

        progress_table = Table.grid()
        progress_table.add_row(
            Panel.fit(
                progress["overall"],
                title="Progress Deletion",
                border_style="magenta",
                padding=(1, 2),
            ),
            Panel.fit(
                progress["counts"],
                title="Counts",
                border_style="bright_black",
                padding=(1, 2),
            ),
        )

        with Live(
            progress_table,
            console=self.console,
            auto_refresh=False,
        ) as live:
            for cur_file_dest in files_dest:
                progress["counts"].advance(tasks_delete["count_files_found"])
                action = conditional_delete(
                    Path(str(cur_file_dest).replace(str(dir_dest), str(dir_source))),
                    cur_file_dest,
                    files_source,
                    dry_run=self.dry_run,
                )
                if action:
                    progress["counts"].advance(tasks_delete["count_files_deleted"])
                    progress["overall"].update(
                        tasks_delete["progress_delete_files"],
                        last_action=cur_file_dest.name,
                    )
                    if self.verbose:
                        live.console.print(
                            f"  - deleting [magenta]'{cur_file_dest.name}'[/] "
                            "from dest dir"
                        )

                progress["overall"].advance(tasks_delete["progress_delete_files"])
                live.refresh()

            for cur_dir_dest in dirs_dest:
                progress["counts"].advance(tasks_delete["count_folders_found"])
                action = conditional_delete(
                    Path(str(cur_dir_dest).replace(str(dir_dest), str(dir_source))),
                    cur_dir_dest,
                    dirs_source,
                    dry_run=self.dry_run,
                )
                if action:
                    progress["counts"].advance(tasks_delete["count_folders_deleted"])
                    progress["overall"].update(
                        tasks_delete["progress_delete_folders"],
                        last_action=cur_dir_dest.name,
                    )
                    if self.verbose:
                        live.console.print(
                            f"  - deleting dir [magenta]'{cur_dir_dest.name}'[/] "
                            "from dest dir"
                        )
                progress["overall"].advance(tasks_delete["progress_delete_folders"])
                live.refresh()



"""Clifs plugin to show disc usage"""


import shutil
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, List, NamedTuple

from clifs import ClifsPlugin
from clifs.utils_cli import cli_bar, set_style, size2str


class UsageInfo(NamedTuple):
    """Disk usage info"""

    total: int
    used: int
    free: int


class DiscUsageExplorer(ClifsPlugin):
    """
    Display disk usage for one or more directories.
    """

    dirs: List[str]

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        parser.add_argument(
            "dirs",
            type=str,
            default=".",
            nargs="*",
            help="Directory or directories do get info from.",
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)
        self._dict_usage: Dict[str, UsageInfo] = self._get_usage_info()

    def run(self) -> None:
        self._print_usage_info()

    def _get_usage_info(self) -> Dict[str, UsageInfo]:
        disc_usage = {}
        for directory in self.dirs:
            disc_usage[directory] = UsageInfo(*shutil.disk_usage(directory))
        return disc_usage

    def _print_usage_info(self) -> None:
        self.console.print()
        for directory, usage_info in self._dict_usage.items():
            name_dir = Path(directory).name if Path(directory).name != "" else directory
            path_dir = str(Path(directory).resolve())
            self.console.print(
                name_dir + "    " + set_style(f"({path_dir})", "bright_black")
            )
            if (frac_used := usage_info.used / usage_info.total) <= 0.7:
                color = "default"
            elif frac_used <= 0.9:
                color = "yellow"
            else:
                color = "red"

            str_total = size2str(usage_info.total, color="default")
            str_used = size2str(usage_info.used, color="default")
            str_free = size2str(usage_info.free, color=color)

            usage_bar = set_style(
                cli_bar(usage_info.used, usage_info.total, print_out=False), color
            )

            self.console.print(
                f"  └── {usage_bar}"
                f"    total: {str_total}"
                f"    used: {str_used}"
                f"    free: {str_free}"
            )



"""Common pytest fixtures"""


import shutil
from pathlib import Path

import pytest

from tests.common.utils_testing import get_files, update_mtime


@pytest.fixture(scope="function")
def dir_testrun(tmp_path: Path):
    # create source dest structure for update test
    shutil.copytree(Path(__file__).parent / "common" / "data", tmp_path / "data")
    return tmp_path


@pytest.fixture(scope="function")
def dirs_source(dir_testrun, num_dirs=2):
    dirs_res = [
        (dir_testrun / "data" / ("dir_source_" + str(i)))
        for i in range(1, num_dirs + 1)
    ]
    # update some files to check mtime consistency
    for dir in dirs_res:
        for file in get_files(dir):
            if "2" in file.stem:
                update_mtime(file)
    return dirs_res


@pytest.fixture(scope="function")
def dirs_dest(dir_testrun, num_dirs=2):
    return [
        (dir_testrun / "data" / ("dir_dest_" + str(i))) for i in range(1, num_dirs + 1)
    ]


@pytest.fixture(scope="function")
def dirs_flatten_dir_source_ref(dir_testrun, num_dirs=2):
    return [
        (dir_testrun / "data" / ("ref_flatten_dir_source_" + str(i)))
        for i in range(1, num_dirs + 1)
    ]


@pytest.fixture(scope="function")
def dirs_keep_dir_source2dest_ref(dir_testrun, num_dirs=2):
    return [
        (dir_testrun / "data" / ("ref_keep_dir_source2dest_" + str(i)))
        for i in range(1, num_dirs + 1)
    ]


@pytest.fixture(scope="function")
def dirs_source_ref(dirs_source):
    # create source reference to check source dir integrity
    dirs_res = []
    for dir in dirs_source:
        dir_res = dir.parent / (dir.name + "_ref")
        shutil.copytree(dir, dir_res)
        dirs_res.append(dir_res)
    return dirs_res


@pytest.fixture(scope="function")
def dirs_dest_ref(dirs_dest):
    # create source reference to check source dir integrity
    dirs_res = []
    for dir in dirs_dest:
        dir_res = dir.parent / (dir.name + "_ref")
        shutil.copytree(dir, dir_res)
        dirs_res.append(dir_res)
    return dirs_res


@pytest.fixture(scope="function")
def dirs_empty(dir_testrun, num_dirs=2):
    lst_dirs = []
    for i in range(num_dirs):
        dir = dir_testrun / "data" / ("dir_empty_" + str(i + 1))
        dir.mkdir()
        lst_dirs.append(dir)
    return lst_dirs


@pytest.fixture(scope="function")
def path_filterlist_txt(dir_testrun):
    return dir_testrun / "data" / "list_filter.txt"


@pytest.fixture(scope="function")
def path_filterlist_csv(dir_testrun):
    return dir_testrun / "data" / "list_filter.csv"


@pytest.fixture(scope="function")
def path_filterlist_tsv(dir_testrun):
    return dir_testrun / "data" / "list_filter.tsv"



"""Clifs plugins for file copying and moving"""

import shutil
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Dict, List

from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, TaskID
from rich.table import Table

from clifs import ClifsPlugin
from clifs.utils_cli import (
    cli_bar,
    get_count_progress,
    get_last_action_progress,
    print_line,
    set_style,
)
from clifs.utils_fs import PathGetterMixin, get_unique_path


class CoMo(ClifsPlugin, PathGetterMixin):
    """
    Base class to copy or move files.

    """

    files2process: List[Path]
    dir_dest: Path
    skip_existing: bool
    keep_all: bool
    flatten: bool
    terse: bool
    dryrun: bool
    move: bool

    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        # add args from FileGetterMixin to arg parser
        super().init_parser_mixin(parser)

        parser.add_argument("dir_dest", type=Path, help="Folder to copy/move files to")
        parser.add_argument(
            "-se",
            "--skip_existing",
            action="store_true",
            help="Do nothing if file already exists in destination "
            "(instead of replacing).",
        )
        parser.add_argument(
            "-ka",
            "--keep_all",
            action="store_true",
            help="Keep both versions if a file already exists in destination "
            "(instead of replacing).",
        )
        parser.add_argument(
            "-flt",
            "--flatten",
            action="store_true",
            help="Flatten folder structure in output directory when running "
            "in recursive mode. "
            "Be careful with files of identical name in different subfolders as "
            "they will overwrite each other by default!",
        )
        parser.add_argument(
            "-t", "--terse", action="store_true", help="Report the summary only."
        )
        parser.add_argument(
            "-dr", "--dryrun", action="store_true", help="Don't touch anything"
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)

        if self.skip_existing and self.keep_all:
            self.console.print(
                "You can only choose to either skip existing files "
                "or keep both versions. Choose wisely!"
            )
            sys.exit(0)

        self.files2process, _ = self.get_paths()

        # define progress
        self.progress: Dict[str, Progress] = {
            "counts": get_count_progress(),
            "overall": get_last_action_progress(),
        }
        self.action = "Moving" if self.move else "Copying"
        self.tasks = self.get_tasks()

        self.progress_table = Table.grid()
        self.progress_table.add_row(
            Panel.fit(
                self.progress["overall"],
                title="Progress",
                border_style="cyan",
                padding=(1, 2),
            ),
            Panel.fit(
                self.progress["counts"],
                title="Counts",
                border_style="bright_black",
                padding=(1, 2),
            ),
        )

    def run(self) -> None:
        self.exit_if_nothing_to_process(self.files2process)
        self.dir_dest.parent.mkdir(exist_ok=True, parents=True)
        self.como()

    def get_tasks(self) -> Dict[str, TaskID]:
        # define overall progress task
        tasks = {
            "progress": self.progress["overall"].add_task(
                f"{self.action} data:  ", total=len(self.files2process), last_action="-"
            ),
        }

        # define counter tasks
        if self.move:
            tasks["files_moved"] = self.progress["counts"].add_task(
                "Files moved:", total=None
            )
        else:
            tasks["files_copied"] = self.progress["counts"].add_task(
                "Files copied:", total=None
            )

        if self.skip_existing:
            tasks["files_skipped"] = self.progress["counts"].add_task(
                "Files skipped:", total=None
            )
        elif self.keep_all:
            tasks["files_renamed"] = self.progress["counts"].add_task(
                "Files renamed:", total=None
            )
        else:
            tasks["files_replaced"] = self.progress["counts"].add_task(
                "Files replaced:", total=None
            )
        return tasks

    def create_file(self, file_src: Path, file_dest: Path) -> None:
        if not self.flatten and not self.dryrun:
            file_dest.parent.mkdir(exist_ok=True, parents=True)
        if self.move:
            if not self.dryrun:
                shutil.move(str(file_src), str(file_dest))
            self.progress["counts"].advance(self.tasks["files_moved"])
        else:
            if not self.dryrun:
                shutil.copy2(str(file_src), str(file_dest))
            self.progress["counts"].advance(self.tasks["files_copied"])

    def como(self) -> None:
        print_line(self.console)
        if self.dryrun:
            print("Dry run:\n")
        self.console.print(
            f"{self.action} {len(self.files2process)} files\n"
            f"from: {self.dir_source}\n"
            f"to:   {self.dir_dest}"
        )

        with Live(
            self.progress_table,
            console=self.console,
            auto_refresh=False,
        ) as live:
            for num_file, file in enumerate(self.files2process, 1):
                skip = False
                txt_report = f"Last: {file.name}"
                filepath_dest = self.get_path_dest(file)
                if filepath_dest.exists():
                    if self.skip_existing:
                        txt_report = set_style(
                            f"Skipped as already present: " f"{file.name}",
                            "warning",
                        )
                        skip = True
                        self.progress["counts"].advance(self.tasks["files_skipped"])
                    elif self.keep_all:
                        filepath_dest_new = get_unique_path(filepath_dest)
                        if filepath_dest_new != filepath_dest:
                            txt_report = set_style(
                                "Changed name as already present: "
                                f"{filepath_dest.name} -> {filepath_dest_new.name}",
                                "warning",
                            )
                            filepath_dest = filepath_dest_new
                            self.progress["counts"].advance(self.tasks["files_renamed"])
                    else:
                        txt_report = set_style(
                            f"Replacing existing version for: {file.name}",
                            "warning",
                        )
                        self.progress["counts"].advance(self.tasks["files_replaced"])

                if not skip:
                    self.create_file(file, filepath_dest)

                last_action = "moved" if self.move else "copied"
                if not self.terse:
                    cli_bar(
                        num_file,
                        len(self.files2process),
                        suffix=f"{last_action}. {txt_report}",
                        console=self.console,
                    )
                self.progress["overall"].update(
                    self.tasks["progress"],
                    last_action=f"{last_action} {file.name}",
                )
                self.progress["overall"].advance(self.tasks["progress"])
                live.refresh()
        print_line(self.console)

    def get_path_dest(self, path_file: Path) -> Path:
        if self.flatten:
            return self.dir_dest / path_file.name
        return Path(str(path_file).replace(str(self.dir_source), str(self.dir_dest)))


class FileMover(CoMo):
    """
    Move files
    """

    def __init__(self, args: Namespace) -> None:
        self.move = True
        super().__init__(args)


class FileCopier(CoMo):
    """
    Copy files
    """

    def __init__(self, args: Namespace) -> None:
        self.move = False
        super().__init__(args)



"""Test the rename plugin"""

from unittest.mock import patch

from clifs.__main__ import main
from tests.common.utils_testing import (
    assert_files_present,
    parametrize_default_ids,
    substr_in_dir_names,
)


@parametrize_default_ids(
    ("pattern", "replacement"),
    [("file", "SUBSTITUTE"), (".txt", "123456"), ("dir", "folder")],
)
@parametrize_default_ids("dirs", [False, True])
def test_rename(dirs_source, dirs_source_ref, pattern, replacement, dirs):
    # run the actual function to test

    for idx_dir, dir in enumerate(dirs_source):
        patch_args = [
            "clifs",
            "ren",
            str(dir),
            "--pattern",
            pattern,
            "--replacement",
            replacement,
            "--recursive",
        ]
        if dirs:
            patch_args.append("--dirs")
        with patch("sys.argv", patch_args), patch("builtins.input", return_value="yes"):
            main()

        print(f"Renaming of {dir.name} went through.")

        if not dirs:
            assert not substr_in_dir_names(dir, sub_str=pattern, files_only=True)
        else:
            assert not substr_in_dir_names(dir, sub_str=pattern, dirs_only=True)

        # revert and check consistency
        patch_args = [
            "clifs",
            "ren",
            str(dir),
            "--pattern",
            replacement,
            "--replacement",
            pattern,
            "--recursive",
            "--skip_preview",
        ]
        if dirs:
            patch_args.append("--dirs")
        with patch("sys.argv", patch_args):
            main()

        print(f"Re-renaming of {dir.name} went through.")

        if not dirs:
            assert not substr_in_dir_names(dir, sub_str=replacement, files_only=True)
        else:
            assert not substr_in_dir_names(dir, sub_str=replacement, dirs_only=True)

        assert_files_present(dir_source=dirs_source_ref[idx_dir], dir_ref=dir)


@parametrize_default_ids("dirs", [False, True])
def test_reject_bad_chars(dirs_source, dirs_source_ref, dirs):
    # run the actual function to test
    bad_char = "~"

    for idx_dir, dir in enumerate(dirs_source):
        patch_args = [
            "clifs",
            "ren",
            str(dir),
            "--pattern",
            "1",
            "--replacement",
            bad_char,
            "--recursive",
        ]
        if dirs:
            patch_args.append("--dirs")
        with patch("sys.argv", patch_args), patch("builtins.input", return_value="yes"):
            main()

        print(f"Renaming of {dir.name} went through.")

        if not dirs:
            assert not substr_in_dir_names(dir, sub_str=bad_char, files_only=True)
        else:
            assert not substr_in_dir_names(dir, sub_str=bad_char, dirs_only=True)

        assert_files_present(dir_source=dirs_source_ref[idx_dir], dir_ref=dir)


@parametrize_default_ids(
    ("dirs", "pattern", "replacement"),
    [(False, "L1_file_1", "L1_file_2"), (True, "subdir_1", "subdir_2")],
)
def test_suffix_duplicates(dirs_source, dirs, pattern, replacement):
    for dir in dirs_source:
        patch_args = [
            "clifs",
            "ren",
            str(dir),
            "--pattern",
            pattern,
            "--replacement",
            replacement,
            "--recursive",
        ]
        if dirs:
            patch_args.append("--dirs")
        with patch("sys.argv", patch_args), patch("builtins.input", return_value="yes"):
            main()

        print(f"Renaming of {dir.name} went through.")

        assert f"{replacement} (2)" in [x.stem for x in dir.rglob("*")]



"""Clifs plugin for regex-based renaming of files and folders"""

import re
import sys
from argparse import ArgumentParser, Namespace

# from collections import Counter
from pathlib import Path
from typing import Counter, List, Literal, Set

from clifs import ClifsPlugin
from clifs.utils_cli import cli_bar, print_line, set_style, user_query
from clifs.utils_fs import INDENT, PathGetterMixin, get_unique_path


class Renamer(ClifsPlugin, PathGetterMixin):
    """
    Regex-based renaming of files and folders.
    """

    pattern: str
    replacement: str
    rename_dirs: bool
    skip_preview: bool

    @classmethod
    def init_parser(cls, parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        # add args from FileGetterMixin to arg parser
        super().init_parser_mixin(parser)

        parser.add_argument(
            "-pt",
            "--pattern",
            default=".*",
            help="Pattern identifying the substring to be replaced. "
            "Supports syntax for `re.sub` from regex module "
            "(https://docs.python.org/3/library/re.html).",
        )
        parser.add_argument(
            "-rp",
            "--replacement",
            default="",
            help="String to use as replacement. "
            "You can use \\1 \\2 etc. to refer to matching groups. "
            "E.g. a pattern like '(.+)\\.(.+)' in combination "
            "with a replacement like '\\1_suffix.\\2' will append suffixes. "
            "Defaults to empty string.",
        )
        parser.add_argument(
            "-d",
            "--dirs",
            dest="rename_dirs",
            action="store_true",
            help="Rename directories instead of files.",
        )
        parser.add_argument(
            "-sp",
            "--skip_preview",
            action="store_true",
            help="Skip preview on what would happen and rename right away. "
            "Only for the brave...",
        )

    def __init__(self, args: Namespace) -> None:
        super().__init__(args)

        self.counter: Counter[str] = Counter()
        self.files, self.dirs = self.get_paths()
        if self.rename_dirs:
            # process deeper folders first to avoid changing paths on the fly
            self.dirs = self.sort_paths(self.dirs)

    def run(self) -> None:
        if not self.rename_dirs:
            if not self.files:
                self.console.print("No files to process")
            else:
                if not self.skip_preview:
                    self.rename(self.files, path_type="files", preview_mode=True)
                    if not user_query(
                        'If you want to apply renaming, give me a "yes" or "y" now!'
                    ):
                        self.console.print("Will not rename for now. See you soon.")
                        sys.exit(0)
                self.rename(self.files, path_type="files", preview_mode=False)
        else:
            if not self.dirs:
                self.console.print("No dirs to process")
            else:
                if not self.skip_preview:
                    self.rename(self.dirs, path_type="dirs", preview_mode=True)
                    if not user_query(
                        'If you want to apply renaming, give me a "yes" or "y" now!'
                    ):
                        self.console.print("Will not rename for now. See you soon.")
                        sys.exit(0)
                self.rename(self.dirs, path_type="dirs", preview_mode=False)

    def rename(
        self,
        paths: List[Path],
        path_type: Literal["files", "dirs"],
        preview_mode: bool = True,
    ) -> None:
        self.counter.clear()
        self.counter["paths_total"] = len(paths)

        self.console.print(f"Renaming {self.counter['paths_total']} {path_type}.")
        print_line(self.console)
        paths_to_be_added: Set[Path] = set()
        paths_to_be_deleted: Set[Path] = set()
        if preview_mode:
            self.console.print("Preview:")

        num_path = 0
        for num_path, path in enumerate(paths, 1):
            name_old = path.name
            name_new = re.sub(self.pattern, self.replacement, name_old)
            message_rename = f"{name_old:35} -> {name_new:35}"

            # skip items if renaming would result in bad characters
            found_bad_chars = self.find_bad_char(name_new)
            if found_bad_chars:
                message_rename += set_style(
                    f"{INDENT}Error: not doing renaming as it would result "
                    f"in bad characters: '{','.join(found_bad_chars)}'",
                    "error",
                )
                self.counter["bad_results"] += 1
                self.print_rename_message(
                    message_rename,
                    num_path,
                    preview_mode=preview_mode,
                )
                continue

            # make sure resulting paths are unique
            path_new = path.parent / name_new
            path_unique = get_unique_path(
                path_new,
                set_taken=paths_to_be_added,
                set_free=paths_to_be_deleted | {path},
            )

            if path_new != path_unique:
                path_new = path_unique
                name_new = path_unique.name
                message_rename = f"{name_old:35} -> {name_new:35}"
                message_rename += set_style(
                    f"{INDENT}Warning: name already exists. Adding number suffix.",
                    "warning",
                )
                self.counter["name_conflicts"] += 1

            # skip items that are not renamed
            if path_new == path:
                message_rename = set_style(message_rename, "bright_black")
                self.print_rename_message(
                    message_rename,
                    num_path,
                    preview_mode=preview_mode,
                )
                continue

            self.print_rename_message(
                message_rename,
                num_path,
                preview_mode=preview_mode,
            )
            if not preview_mode:
                path.rename(path_new)
                self.counter["paths_renamed"] += 1
            else:
                paths_to_be_added.add(path_new)
                if path_new in paths_to_be_deleted:
                    paths_to_be_deleted.remove(path_new)
                paths_to_be_deleted.add(path)

        if self.counter["bad_results"] > 0:
            self.console.print(
                set_style(
                    f"Warning: {self.counter['bad_results']} out of "
                    f"{self.counter['paths_total']} files not renamed as it would "
                    "result in bad characters.",
                    "warning",
                )
            )

        if self.counter["name_conflicts"] > 0:
            self.console.print(
                set_style(
                    f"Warning: {self.counter['name_conflicts']} out of "
                    f"{self.counter['paths_total']} renamings would have resulted in "
                    "name conflicts. Added numbering suffixes to get unique names.",
                    "warning",
                )
            )

        if not preview_mode:
            self.console.print(
                f"Hurray, {num_path} {path_type} have been processed, "
                f"{self.counter['paths_renamed']} have been renamed."
            )
        print_line(self.console)

    def print_rename_message(
        self,
        message: str,
        num_file: int,
        preview_mode: bool = False,
        space_prefix: str = "    ",
    ) -> None:
        if preview_mode:
            self.console.print(space_prefix + message)
        else:
            cli_bar(
                num_file,
                self.counter["paths_total"],
                suffix=space_prefix + message,
                console=self.console,
            )

    @staticmethod
    def find_bad_char(string: str) -> List[str]:
        """Check stings for characters causing problems in windows file system."""
        bad_chars = r"~“#%&*:<>?/\{|}"
        return [x for x in bad_chars if x in string]



"""Utilities for the command line interface"""

from enum import Enum
from typing import Iterable, Optional

from rich.console import Console, RenderableType
from rich.progress import BarColumn, Progress, TaskProgressColumn, TimeRemainingColumn
from rich.rule import Rule
from rich.theme import Theme

THEME_RICH = Theme(
    {
        "bar.complete": "default",
        "bar.finished": "green",
        "bar.back": "bright_black",
        "progress.percentage": "default",
        "progress.remaining": "bright_black",
        "rule.line": "default",
        "warning": "yellow",
        "error": "red",
        "folder": "yellow",
    },
)
CONSOLE = Console(theme=THEME_RICH, highlight=False)


class AnsiColor(str, Enum):
    """Ansi escape colors."""

    DEFAULT = "\033[0m"
    BLUE = "\033[0;34m"
    CYAN = "\033[0;36m"
    GRAY = "\033[0;90m"
    GREEN = "\033[0;32m"
    MAGENTA = "\033[0;35m"
    RED = "\033[0;31m"
    WHITE = "\033[0;37m"
    YELLOW = "\033[0;93m"

    def __str__(self) -> str:
        return self.value


def set_style(
    string: str,
    style: str = "red",
) -> str:
    """Set rich style for strings.

    :param string: String
    :param style: Style to set, defaults to "red"
    :return: String wrapped in style markup
    """
    if string.endswith("\\") and not string.endswith("\\\\"):
        string += "\\"
    return f"[{style}]{string}[/{style}]"


class LastActionProgress(Progress):
    """Progress showing the last action in a separate line."""

    def get_renderables(self) -> Iterable[RenderableType]:
        """Get a number of renderables for the progress display."""

        table = self.make_tasks_table(self.tasks)
        yield table
        for task in self.tasks:
            yield (
                f"{task.fields.get('last_action_desc','Last action')}: "
                f"{task.fields.get('last_action', '-')}"
            )


def size2str(size: float, color: str = "cyan") -> str:
    """Format data size in bites to nicely readable units.

    :param size: Input size in bites
    :param color: Output color for rich markup, defaults to "cyan"
    :return: String of data size wrapped in color markup
    """
    if size < 1024**2:
        unit = "KB"
        size = round(size / 1024, 2)
    elif size < 1024**3:
        unit = "MB"
        size = round(size / 1024**2, 2)
    elif size < 1024**4:
        unit = "GB"
        size = round(size / 1024**3, 2)
    elif size < 1024**5:
        unit = "TB"
        size = round(size / 1024**4, 2)
    else:
        unit = "PB"
        size = round(size / 1024**5, 2)
    return f"[{color}]{size:7.2f} {unit}[/{color}]"


def cli_bar(  # pylint: disable=too-many-arguments
    status: int,
    total: int,
    suffix: str = "",
    print_out: bool = True,
    bar_len: int = 20,
    console: Optional[Console] = None,
) -> str:
    """Create progress bar ans either print directly to console or return as string.

    :param status: Number of finished steps
    :param total: Total number of expected steps
    :param suffix: Suffix to add to the bar, defaults to ""
    :param print_out: Whether to print directly or not, defaults to True
    :param bar_len: Number of characters used for the bar, defaults to 20
    :param console: rich.console object to print to, defaults to None
    :return: Progress bar including percent indication and the suffix
    """
    filled_len = int(round(bar_len * status / float(total)))
    percents = round(100.0 * status / float(total), 1)
    res_bar = "█" * filled_len + "-" * (bar_len - filled_len)
    output = f"|{res_bar}| {percents:5}% {suffix}"
    if print_out:
        if console:
            console.print(output)
        else:
            print(output)
    return output


def user_query(message: str) -> bool:
    """Run user query.

    :param message: Message to show
    :return: Boolean indicating if the user input was "y" or "yes"
    """
    yes = {"yes", "y"}
    print(message)
    choice = input().lower()
    return choice in yes


def print_line(console: Console = CONSOLE, title: str = "") -> None:
    """Print a line to the console.

    :param console: rich.console to print to, defaults to CONSOLE
    :param title: Title included in the line, defaults to ""
    """
    console.print(Rule(title=title, align="left"))


def get_count_progress() -> Progress:
    """Get instance of a count progress.

    :return: progress
    """
    return Progress(
        "{task.description}",
        "{task.completed}",
    )


def get_last_action_progress() -> LastActionProgress:
    """Get instance of  a progress bar indicating the last action in a separate line.

    :return: Progress
    """
    return LastActionProgress(
        "{task.description}",
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(),
    )



"""Test the tree plugin"""


import re
from argparse import Namespace
from unittest.mock import patch

import pytest

from clifs.__main__ import main
from clifs.plugins.tree import DirectoryTree
from tests.common.utils_testing import escape_rich_style, parametrize_default_ids


@pytest.fixture()
def trees_source_dir():
    dir_1_tree = (
        "dir_source_1    7.00 KB\n"
        "├── L1_file_1.txt    1.00 KB\n"
        "├── L1_file_2.txt    1.00 KB\n"
        "├── subdir_1    4.00 KB\n"
        "│   ├── L2_file_1.txt    1.00 KB\n"
        "│   └── subsubdir_1    3.00 KB\n"
        "│       ├── L3_file_1.txt    1.00 KB\n"
        "│       ├── L3_file_2.txt    1.00 KB\n"
        "│       └── L3_file_3.txt    1.00 KB\n"
        "└── subdir_2    1.00 KB\n"
        "    └── L2_file_2.txt    1.00 KB\n"
    )

    dir_2_tree = (
        "dir_source_2    8.00 KB\n"
        "├── L1_file_1.txt    1.00 KB\n"
        "├── L1_file_2.txt    1.00 KB\n"
        "├── L1_file_3.txt    1.00 KB\n"
        "├── subdir_1    4.00 KB\n"
        "│   ├── L2_file_1.txt    1.00 KB\n"
        "│   └── subsubdir_1    3.00 KB\n"
        "│       ├── L3_file_1.txt    1.00 KB\n"
        "│       ├── L3_file_2.txt    1.00 KB\n"
        "│       └── L3_file_3.txt    1.00 KB\n"
        "└── subdir_2    1.00 KB\n"
        "    └── L2_file_2.txt    1.00 KB\n"
    )
    return (dir_1_tree, dir_2_tree)


@parametrize_default_ids("dirs_only", [False, True])
@parametrize_default_ids("hide_sizes", [False, True])
@parametrize_default_ids("depth", [None, 1, 2])
def test_tree(dirs_source, trees_source_dir, dirs_only, hide_sizes, depth):
    for idx, folder in enumerate(dirs_source):
        with patch("os.stat_result.st_size", 1024):
            tree = DirectoryTree(
                Namespace(
                    root_dir=folder,
                    dirs_only=dirs_only,
                    hide_sizes=hide_sizes,
                    depth=depth,
                )
            )
        tree.run()
        print(f"Tree generation for {folder.name} went through.")

        exp_tree = trees_source_dir[idx].splitlines()
        if depth is not None:
            for idx, line in enumerate(exp_tree.copy()):
                if match := re.match(r".*L(\d+)_file.*", line):
                    exp_depth = int(match[1])
                else:
                    exp_depth = len(re.findall("sub", line))
                if exp_depth > depth:
                    exp_tree.remove(line)
        if dirs_only:
            for line in exp_tree.copy():
                if "dir" not in line:
                    exp_tree.remove(line)
        if hide_sizes:
            exp_tree_no_sizes = []
            for line in exp_tree:
                exp_tree_no_sizes.append(line[:-9].rstrip())
            exp_tree = exp_tree_no_sizes

        act_tree = tree.__str__().splitlines()
        for lineidx, line in enumerate(act_tree):
            try:
                assert escape_rich_style(line) == exp_tree[lineidx].replace(
                    "\n", ""
                ), f"Trees for {folder} are not identical."
            except AssertionError:
                print(f"{exp_tree=}")
                print(f"{act_tree=}")
                raise


def test_entry(dirs_source, capfd, trees_source_dir):
    for dir, exp_tree in zip(dirs_source, trees_source_dir):
        # run the actual function to test
        with patch("sys.argv", ["clifs", "tree", str(dir)]), patch(
            "os.stat_result.st_size", 1024
        ):
            main()
        out, _ = capfd.readouterr()
        for line in exp_tree.splitlines():
            assert line in out, (
                "Expected the following line in the output but did not find it:\n"
                f"{line}"
            )



"""Utilities for the file system"""

import csv
import re
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple

INDENT = "    "


class PathGetterMixin:
    """
    Get paths from a source directory by different filter methods.
    """

    dir_source: Path
    recursive: bool
    filterlist: Path
    filterlistheader: str
    filterlistsep: str
    filterstring: str

    @staticmethod
    def init_parser_mixin(parser: ArgumentParser) -> None:
        """
        Adding arguments to an argparse parser. Needed for all clifs_plugins.
        """
        parser.add_argument(
            "dir_source",
            type=Path,
            help="Folder with files to copy/move from",
        )
        parser.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help="Search recursively in source folder",
        )
        parser.add_argument(
            "-fl",
            "--filterlist",
            default=None,
            type=Path,
            help="Path to a txt or csv file containing a list of files to process. "
            "In case of a CSV, separator and header can be provided additionally via "
            "the parameters `filterlistsep` and `filterlistheader`. "
            "If no header is provided, each line in the file is treated as individual "
            "file name.",
        )
        parser.add_argument(
            "-flh",
            "--filterlistheader",
            default=None,
            help="Header of the column to use as filter "
            "from a csv provided as filterlist."
            " If no header is provided, "
            "each line in the file is read as individual item name.",
        )
        parser.add_argument(
            "-fls",
            "--filterlistsep",
            default=",",
            help="Separator to use for csv provided as filter list. Default: ','",
        )
        parser.add_argument(
            "-fs",
            "--filterstring",
            default=None,
            help="Substring identifying files to be copied. not case sensitive.",
        )

    def get_paths(self) -> Tuple[List[Path], List[Path]]:
        """Get file and folder paths depending on set filters

        :return: Lists of file paths and folder paths matching the filters respectively
        """
        files, dirs = self._get_paths_by_filterstring(
            self.dir_source, filterstring=self.filterstring, recursive=self.recursive
        )

        if self.filterlist:
            list_filter = self._list_from_csv()
            files = [i for i in files if i.name in list_filter]
            dirs = [i for i in dirs if i.name in list_filter]
        return files, dirs

    @staticmethod
    def exit_if_nothing_to_process(items: List[Any]) -> None:
        """Exit running process if list of files to process is empty"""
        if not items:
            print("Nothing to process.")
            sys.exit(0)

    @staticmethod
    def sort_paths(paths: List[Path]) -> List[Path]:
        """Sort by inverse depth and str

        :param paths: List of paths to sort
        :return: Sorted list of paths
        """
        return sorted(paths, key=lambda x: (-len(x.parents), str(x)))

    @staticmethod
    def _get_paths_by_filterstring(
        dir_source: Path, filterstring: Optional[str] = None, recursive: bool = False
    ) -> Tuple[List[Path], List[Path]]:
        """Get files by substring filter on the file name.

        :param dir_source: directory to search for files in
        :param filterstring: Substring that must be included in a file name.
            If set to None, files are not filtered by substring. Defaults to None.
        :param recursive: Search recursively, defaults to False
        :return: Lists of file paths and dir paths matching the filter respectively
        """
        pattern_search = f"*{filterstring}*" if filterstring else "*"
        if recursive:
            pattern_search = "**/" + pattern_search
        files = []
        dirs = []
        for path in dir_source.glob(pattern_search):
            if path.is_dir():
                dirs.append(path.resolve())
            else:
                files.append(path.resolve())

        return files, dirs

    def _list_from_csv(self) -> List[str]:
        if not self.filterlistheader:
            res_list = self.filterlist.open().read().splitlines()
        else:
            with self.filterlist.open(newline="") as infile:
                reader = csv.DictReader(infile, delimiter=self.filterlistsep)
                res_list = []
                for row in reader:
                    try:
                        res_list.append(row[self.filterlistheader])
                    except KeyError:
                        print(
                            "Provided csv does not contain header "
                            f"'{self.filterlistheader}'. Found headers:\n"
                            f"{list(row.keys())}"
                        )
                        raise
        return res_list


def get_unique_path(
    path_candidate: Path,
    set_taken: Optional[Set[Path]] = None,
    set_free: Optional[Set[Path]] = None,
) -> Path:
    """Given a name candidate get a unique file name in a given directory.

    Adds number suffixes in form ' (#)' if file name is already taken.

    :param path_candidate: Candidate for a file path.
    :param set_taken: Optional set of additional paths which are considered as already
        taken, defaults to None
    :param set_free: Optional sets of paths that are considered as not taken even if
        corresponding files exist, defaults to None
    :raises ValueError: If there are common elements in 'set_taken' and 'set_free'
    :return: Unique file path
    """
    if set_taken is None:
        set_taken = set()
    if set_free is None:
        set_free = set()
    if intersect := set_taken.intersection(set_free):
        raise ValueError(
            "Params 'set_taken' and 'set_free' contain common elements: \n"
            f"{intersect=}."
        )

    path_new = path_candidate
    if (path_new.exists() or path_new in set_taken) and (path_new not in set_free):
        name_file = path_new.stem
        count_match = re.match(r".* \((\d+)\)$", name_file)
        if count_match:
            count = int(count_match.group(1)) + 1
            name_file = " ".join(name_file.split(" ")[0:-1])
        else:
            count = 2

        while (path_new.exists() or path_new in set_taken) and (
            path_new not in set_free
        ):
            name_file_new = name_file + f" ({count})"
            path_new = path_candidate.parent / (name_file_new + path_candidate.suffix)
            count += 1
    return path_new
